use alloc::vec;
use alloc::vec::Vec;
use vapfs::{BlockDevice, Index};
use crate::{read_datablock, write_datablock};
use crate::structs::{ListBlock, Superblock};

pub struct ListblockIter<'a> {
    listblock: &'a mut ListBlock,
    index: usize,
    buf1: (Index, Vec<u8>),
    buf2: (Index, Vec<u8>),
    buf3: (Index, Vec<u8>),
    buf4: (Index, Vec<u8>),
    buf5: (Index, Vec<u8>),
    buf6: (Index, Vec<u8>),
}

impl<'a> ListblockIter<'a> {
    pub fn new(listblock: &'a mut ListBlock) -> Self {
        Self { listblock, index: 0, buf1: (0, vec![]), buf2: (0, vec![]), buf3: (0, vec![]), buf4: (0, vec![]), buf5: (0, vec![]), buf6: (0, vec![]) }
    }

    pub fn next(&mut self, sb: &Superblock, bd: &mut dyn BlockDevice) -> Option<Index> {
        if self.index as Index >= self.listblock.count {
            return None;
        }

        if self.index < 32 {
            let index = self.listblock.direct_block_addresses[self.index];
            self.index += 1;
            return Some(index);
        } else {
            // copy pasted from structs.rs
            // if greater than 32, see the following:
            // let N = (maximum number of pointers in an indirect block) * 32
            // (beginning...count)
            // 32..N: single indirect block
            // 32+N..N^2: double indirect block
            // 32+N^2..N^3: triple indirect block
            // 32+N^3..N^4: quadruple indirect block
            // 32+N^4..N^5: quintuple indirect block
            // 32+N^5..N^6: sextuple indirect block

            // block index is address / (max_per_block ^ (1 if single, 2 if double, etc))
            // after getting your layer's address, repeat the previous layer's process

            let max_per_block = sb.block_size as u64 / 8;
            let N = max_per_block * 32;
            let N2 = N * N;
            let N3 = N2 * N;
            let N4 = N3 * N;
            let N5 = N4 * N;
            let N6 = N5 * N;

            let mut address = self.index as u64 - 32;

            // todo: you could probably rewrite this using like recursion or smth

            match address {
                _ if address < N => {
                    let block_index = address / max_per_block;
                    let address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != self.listblock.single_indirect_block_address[block_index as usize] {
                        self.buf1.1 = read_datablock(self.listblock.single_indirect_block_address[block_index as usize], sb, bd);
                        self.buf1.0 = self.listblock.single_indirect_block_address[block_index as usize];
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(address * 8) as usize..(address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N2 => {
                    let address = address - N;
                    let block_index = (address) / (max_per_block ^ 2);
                    if self.buf2.1.len() == 0 || self.buf2.0 != self.listblock.double_indirect_block_address[block_index as usize] {
                        self.buf2.1 = read_datablock(self.listblock.double_indirect_block_address[block_index as usize], sb, bd);
                        self.buf2.0 = self.listblock.double_indirect_block_address[block_index as usize];
                    }
                    let layer2_address = address / max_per_block;
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N3 => {
                    let address = address - N2;
                    let block_index = (address) / (max_per_block ^ 3);
                    if self.buf3.1.len() == 0 || self.buf3.0 != self.listblock.triple_indirect_block_address[block_index as usize] {
                        self.buf3.1 = read_datablock(self.listblock.triple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf3.0 = self.listblock.triple_indirect_block_address[block_index as usize];
                    }
                    let layer3_address = address / (max_per_block ^ 2);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N4 => {
                    let address = address - N3;
                    let block_index = (address) / (max_per_block ^ 4);
                    if self.buf4.1.len() == 0 || self.buf4.0 != self.listblock.quadruple_indirect_block_address[block_index as usize] {
                        self.buf4.1 = read_datablock(self.listblock.quadruple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf4.0 = self.listblock.quadruple_indirect_block_address[block_index as usize];
                    }
                    let layer4_address = address / (max_per_block ^ 3);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf4.1[(layer4_address * 8) as usize..(layer4_address * 8 + 8) as usize]);
                    let layer3_block = u64::from_be_bytes(buf);
                    let layer3_address = (address % (max_per_block ^ 3)) / (max_per_block ^ 2);
                    if self.buf3.1.len() == 0 || self.buf3.0 != layer3_block {
                        self.buf3.1 = read_datablock(layer3_block, sb, bd);
                        self.buf3.0 = layer3_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N5 => {
                    let address = address - N4;
                    let block_index = (address) / (max_per_block ^ 5);
                    if self.buf5.1.len() == 0 || self.buf5.0 != self.listblock.quintuple_indirect_block_address[block_index as usize] {
                        self.buf5.1 = read_datablock(self.listblock.quintuple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf5.0 = self.listblock.quintuple_indirect_block_address[block_index as usize];
                    }
                    let layer5_address = address / (max_per_block ^ 4);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf5.1[(layer5_address * 8) as usize..(layer5_address * 8 + 8) as usize]);
                    let layer4_block = u64::from_be_bytes(buf);
                    let layer4_address = (address % (max_per_block ^ 4)) / (max_per_block ^ 3);
                    if self.buf4.1.len() == 0 || self.buf4.0 != layer4_block {
                        self.buf4.1 = read_datablock(layer4_block, sb, bd);
                        self.buf4.0 = layer4_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf4.1[(layer4_address * 8) as usize..(layer4_address * 8 + 8) as usize]);
                    let layer3_block = u64::from_be_bytes(buf);
                    let layer3_address = (address % (max_per_block ^ 3)) / (max_per_block ^ 2);
                    if self.buf3.1.len() == 0 || self.buf3.0 != layer3_block {
                        self.buf3.1 = read_datablock(layer3_block, sb, bd);
                        self.buf3.0 = layer3_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N6 => {
                    let address = address - N5;
                    let block_index = (address) / (max_per_block ^ 6);
                    if self.buf6.1.len() == 0 || self.buf6.0 != self.listblock.sextuple_indirect_block_address[block_index as usize] {
                        self.buf6.1 = read_datablock(self.listblock.sextuple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf6.0 = self.listblock.sextuple_indirect_block_address[block_index as usize];
                    }
                    let layer6_address = address / (max_per_block ^ 5);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf6.1[(layer6_address * 8) as usize..(layer6_address * 8 + 8) as usize]);
                    let layer5_block = u64::from_be_bytes(buf);
                    let layer5_address = (address % (max_per_block ^ 5)) / (max_per_block ^ 4);
                    if self.buf5.1.len() == 0 || self.buf5.0 != layer5_block {
                        self.buf5.1 = read_datablock(layer5_block, sb, bd);
                        self.buf5.0 = layer5_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf5.1[(layer5_address * 8) as usize..(layer5_address * 8 + 8) as usize]);
                    let layer4_block = u64::from_be_bytes(buf);
                    let layer4_address = (address % (max_per_block ^ 4)) / (max_per_block ^ 3);
                    if self.buf4.1.len() == 0 || self.buf4.0 != layer4_block {
                        self.buf4.1 = read_datablock(layer4_block, sb, bd);
                        self.buf4.0 = layer4_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf4.1[(layer4_address * 8) as usize..(layer4_address * 8 + 8) as usize]);
                    let layer3_block = u64::from_be_bytes(buf);
                    let layer3_address = (address % (max_per_block ^ 3)) / (max_per_block ^ 2);
                    if self.buf3.1.len() == 0 || self.buf3.0 != layer3_block {
                        self.buf3.1 = read_datablock(layer3_block, sb, bd);
                        self.buf3.0 = layer3_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ => None,
            }
        }
    }

    /// returns the amount of extra datablocks needed to be able to append the given amount of datablocks
    /// returns None if an error occurs
    pub fn needed_to_append(&mut self, sb: &Superblock, bd: &mut dyn BlockDevice, want_to_append: u64) -> Option<Index> {
        let max_per_block = sb.block_size as u64 / 8; // maximum amount of pointers we can store per datablock
        let N = max_per_block * 32; // maximum amount of pointers we can store in the single indirect blocks section of a listblock
        let N2 = N * N; // maximum amount of pointers we can store in the double indirect blocks section of a listblock
        let N3 = N2 * N; // you get the idea
        let N4 = N3 * N;
        let N5 = N4 * N;
        let N6 = N5 * N;

        let mut indirect_blocks_needed = 0;
        // if we can fit it in the direct blocks, no extra blocks needed
        let dba_count = self.listblock.count + want_to_append; // how many blocks of data we will have after appending
        if dba_count < 32 {
            return Some(0);
        } else if dba_count < N { // if we can fit it in the single indirect blocks
            // yes we can
            // how many indirect blocks do we need?
            indirect_blocks_needed = (dba_count / max_per_block) + 1;
        } else if dba_count < N2 {
            // no, but we can fit it in the double indirect blocks
            // first, fill up the single indirect blocks
            indirect_blocks_needed = N / max_per_block;
            let datablocks_left = dba_count - N;
            // how many double indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N) + 1;
            // how many single indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / max_per_block) + 1;
        } else if dba_count < N3 {
            // this fills up the single and double indirect blocks
            indirect_blocks_needed = N / max_per_block; // 32 single indirect blocks
            indirect_blocks_needed += N2 / N;
            let datablocks_left = dba_count - N2;
            // how many triple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N2) + 1;
            // how many double indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N) + 1;
            // how many single indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / max_per_block) + 1;
        } else if dba_count < N4 {
            // this fills up the single, double, and triple indirect blocks
            indirect_blocks_needed = N / max_per_block;
            indirect_blocks_needed += N2 / N;
            indirect_blocks_needed += N3 / N2;
            let datablocks_left = dba_count - N3;
            // how many quadruple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N3) + 1;
            // how many triple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N2) + 1;
            // how many double indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N) + 1;
            // how many single indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / max_per_block) + 1;
        } else if dba_count < N5 {
            // this fills up the single, double, triple, and quadruple indirect blocks
            indirect_blocks_needed = N / max_per_block;
            indirect_blocks_needed += N2 / N;
            indirect_blocks_needed += N3 / N2;
            indirect_blocks_needed += N4 / N3;
            let datablocks_left = dba_count - N4;
            // how many quintuple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N4) + 1;
            // how many quadruple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N3) + 1;
            // how many triple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N2) + 1;
            // how many double indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N) + 1;
            // how many single indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / max_per_block) + 1;
        } else if dba_count < N6 {
            // this fills up the single, double, triple, quadruple, and quintuple indirect blocks
            indirect_blocks_needed = N / max_per_block;
            indirect_blocks_needed += N2 / N;
            indirect_blocks_needed += N3 / N2;
            indirect_blocks_needed += N4 / N3;
            indirect_blocks_needed += N5 / N4;
            let datablocks_left = dba_count - N5;
            // how many sextuple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N5) + 1;
            // how many quintuple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N4) + 1;
            // how many quadruple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N3) + 1;
            // how many triple indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N2) + 1;
            // how many double indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / N) + 1;
            // how many single indirect blocks do we need?
            indirect_blocks_needed += (datablocks_left / max_per_block) + 1;
        } else {
            // congratulations, you've attempted to write around 2105000 terabytes of data
            return None;
        }

        if want_to_append != 0 {
            // subtract the amount of indirect blocks we already have
            indirect_blocks_needed -= self.needed_to_append(sb, bd, 0)?;
        }

        Some(indirect_blocks_needed)
    }

    /// appends the given datablocks to the list block, using the provided extra indirect blocks
    /// this function is unsafe because it assumes that you've calculated the correct amount of indirect blocks needed
    /// also, no journaling is done to the writes and it is assumed that you are working with scratch data and not live data
    /// returns true on success, false on failure
    pub unsafe fn append(&mut self, sb: &Superblock, bd: &mut dyn BlockDevice, to_append: &[Index], extra: &mut Vec<Index>) -> bool {
        let mut i = 0;
        let mut taken = 0;
        let mut skip_data = self.listblock.count as u64;

        let max_per_block = sb.block_size as u64 / 8; // maximum amount of pointers we can store per datablock
        let N = max_per_block * 32; // maximum amount of pointers we can store in the single indirect blocks section of a listblock
        let N2 = N * N; // maximum amount of pointers we can store in the double indirect blocks section of a listblock
        let N3 = N2 * N; // you get the idea
        let N4 = N3 * N;
        let N5 = N4 * N;
        let N6 = N5 * N;

        // fill first 32 direct if possible
        if skip_data < 32 {
            for block in 0..32 {
                if i >= to_append.len() as u64 {
                    break;
                }
                if skip_data > 0 {
                    skip_data -= 1;
                    continue;
                }
                self.listblock.direct_block_addresses[block as usize] = to_append[i as usize];
                i += 1;
                taken += 1;
            }
        }

        fn fillwithdata_1(
            sb: &Superblock, bd: &mut dyn BlockDevice, i: &mut u64, skip_data: &mut u64, data: &[Index],
            max_per_block: u64, indirect_blocks: &mut Vec<Index>,
            taken: &mut usize, dbas: &mut [Index], max: usize) -> bool {
            for block1 in 0..max {
                if *i >= data.len() as u64 {
                    break; // don't crash if the programmer fucked up
                }

                let mut list = vec![0u64; max_per_block as usize];
                if *skip_data < max_per_block {
                    // we will write new data within this loop iteration, copy data from disk into
                    // list
                    let mut buf = read_datablock(dbas[block1], sb, bd);
                    for k in 0..*skip_data {
                        let mut addr: [u8; 8] = [0; 8];
                        addr.copy_from_slice(&buf[(k * 8) as usize..(k * 8 + 8) as usize]);
                        let addr = u64::from_be_bytes(addr);
                        list[k as usize] = addr;
                    }
                }
                let mut j = 0;
                while j < max_per_block { // for each space in an indirect block
                    let index = *i % max_per_block;
                    if *i >= data.len() as u64 {
                        break;
                    }
                    if *skip_data > 0 {
                        *skip_data -= 1;
                        continue;
                    }
                    list[index as usize] = data[*i as usize].to_be();
                    *i += 1;
                    j += 1;
                }

                if *skip_data > 0 {
                    // no data should've been written, skip this block
                    continue;
                }

                let buf = list.iter().map(|x| x.to_be_bytes()).flatten().collect::<Vec<u8>>();
                if !unsafe { write_datablock(indirect_blocks[*taken], sb, bd, &buf) } {
                    return false;
                }
                *taken += 1;
                dbas[block1] = indirect_blocks[*taken - 1];
            }

            true
        }

        // single indirect blocks (should write N datablocks)
        if !fillwithdata_1(sb, bd, &mut i, &mut skip_data, to_append, max_per_block,
                           extra, &mut taken,
                           &mut self.listblock.single_indirect_block_address, 32) {
            return false;
        }

        // double indirect blocks (should write N^2 datablocks)
        for block2 in 0..32 {
            if i >= to_append.len() as u64 {
                break;
            }
            let mut list = vec![0u64; max_per_block as usize];

            //

            if !fillwithdata_1(sb, bd, &mut i, &mut skip_data, to_append, max_per_block,
                               extra, &mut taken,
                               &mut list, max_per_block as usize) {
                return false;
            }
            let buf = list.iter().map(|x| x.to_be_bytes()).flatten().collect::<Vec<u8>>();
            if skip > 0 {
                skip -= 1;
                continue;
            }
            if !unsafe { write_datablock(extra[taken], sb, bd, &buf) } {
                return false;
            }
            taken += 1;
            self.listblock.double_indirect_block_address[block2 as usize] = extra[taken - 1];
        }

        // triple indirect blocks
        fn fillwithdata_2(
            sb: &Superblock, bd: &mut dyn BlockDevice, i: &mut u64, skip: &mut u64, data: &[Index],
            block_size: usize, max_per_block: u64, indirect_blocks: &mut Vec<Index>,
            taken: &mut usize, dbas: &mut [Index], max: usize) -> bool {
            for block3 in 0..32 { // triple
                if *i >= data.len() as u64 {
                    break;
                }
                let mut buf = vec![0u8; sb.block_size as usize];
                for block2 in 0..max_per_block { // double
                    if *i >= data.len() as u64 {
                        break;
                    }
                    let mut buf2 = vec![0u64; max_per_block as usize];
                    // single
                    fillwithdata_1(sb, bd, i, skip, data, block_size, max_per_block,
                                   indirect_blocks, taken,
                                   &mut buf2, max_per_block as usize);
                    let buf2 = buf2.iter().map(|x| x.to_be_bytes()).flatten().collect::<Vec<u8>>();
                    if *skip > 0 {
                        *skip -= 1;
                        continue;
                    }
                    if !unsafe { write_datablock(indirect_blocks[*taken], sb, bd, &buf2) } {
                        return false;
                    }
                    *taken += 1;
                    buf[block2 as usize * 8..(block2 as usize + 1) * 8].copy_from_slice(&indirect_blocks[*taken - 1].to_be_bytes());
                }
                if *skip > 0 {
                    *skip -= 1;
                    continue;
                }
                if !unsafe { write_datablock(indirect_blocks[*taken], sb, bd, &buf) } {
                    return false;
                }
                *taken += 1;
                dbas[block3 as usize] = indirect_blocks[*taken - 1];
            }
            true
        }
        fillwithdata_2(sb, bd, &mut i, &mut skip, to_append, sb.block_size as usize, max_per_block,
                       extra, &mut taken,
                       &mut self.listblock.triple_indirect_block_address, 32);

        // quadruple indirect blocks
        for block4 in 0..32 {
            if i >= to_append.len() as u64 {
                break;
            }
            let mut list = vec![0u64; max_per_block as usize];
            fillwithdata_2(sb, bd, &mut i, &mut skip, to_append, sb.block_size as usize, max_per_block,
                           extra, &mut taken,
                           &mut list, max_per_block as usize);
            let buf = list.iter().map(|x| x.to_be_bytes()).flatten().collect::<Vec<u8>>();
            if skip > 0 {
                skip -= 1;
                continue;
            }
            if !unsafe { write_datablock(extra[taken], sb, bd, &buf) } {
                return false;
            }
            taken += 1;
            self.listblock.quadruple_indirect_block_address[block4 as usize] = extra[taken - 1];
        }

        // quintuple indirect blocks
        for block5 in 0..32 {
            if i >= to_append.len() as u64 {
                break;
            }
            let mut buf = vec![0u8; sb.block_size as usize];
            for block4 in 0..max_per_block {
                if i >= to_append.len() as u64 {
                    break;
                }
                let mut list = vec![0u64; max_per_block as usize];
                fillwithdata_2(sb, bd, &mut i, &mut skip, to_append, sb.block_size as usize, max_per_block,
                               extra, &mut taken,
                               &mut list, max_per_block as usize);
                let buf2 = list.iter().map(|x| x.to_be_bytes()).flatten().collect::<Vec<u8>>();
                if skip > 0 {
                    skip -= 1;
                    continue;
                }
                if !unsafe { write_datablock(extra[taken], sb, bd, &buf2) } {
                    return false;
                }
                taken += 1;
                buf[block4 as usize * 8..(block4 as usize + 1) * 8].copy_from_slice(&extra[taken - 1].to_be_bytes());
            }
            if skip > 0 {
                skip -= 1;
                continue;
            }
            if !unsafe { write_datablock(extra[taken], sb, bd, &buf) } {
                return false;
            }
            taken += 1;
            self.listblock.quintuple_indirect_block_address[block5 as usize] = extra[taken - 1];
        }

        // sextuple indirect blocks
        for block6 in 0..32 {
            if i >= to_append.len() as u64 {
                break;
            }
            let mut buf_outer = vec![0u8; max_per_block as usize];
            for block5 in 0..max_per_block {
                if i >= to_append.len() as u64 {
                    break;
                }
                let mut buf = vec![0u8; max_per_block as usize];
                for block4 in 0..max_per_block {
                    if i >= to_append.len() as u64 {
                        break;
                    }
                    let mut list = vec![0u64; max_per_block as usize];
                    fillwithdata_2(sb, bd, &mut i, &mut skip, to_append, sb.block_size as usize, max_per_block,
                                   extra, &mut taken,
                                   &mut list, max_per_block as usize);
                    let buf2 = list.iter().map(|x| x.to_be_bytes()).flatten().collect::<Vec<u8>>();
                    if skip > 0 {
                        skip -= 1;
                        continue;
                    }
                    if !unsafe { write_datablock(extra[taken], sb, bd, &buf2) } {
                        return false;
                    }
                    taken += 1;
                    buf[block4 as usize * 8..(block4 as usize + 1) * 8].copy_from_slice(&extra[taken - 1].to_be_bytes());
                }
                if skip > 0 {
                    skip -= 1;
                    continue;
                }
                if !unsafe { write_datablock(extra[taken], sb, bd, &buf) } {
                    return false;
                }
                taken += 1;
                buf_outer[block5 as usize * 8..(block5 as usize + 1) * 8].copy_from_slice(&extra[taken - 1].to_be_bytes());
            }
            if skip > 0 {
                skip -= 1;
                continue;
            }
            if !unsafe { write_datablock(extra[taken], sb, bd, &buf_outer) } {
                return false;
            }
            taken += 1;
            self.listblock.sextuple_indirect_block_address[block6 as usize] = extra[taken - 1];

        }

        todo!()
    }
}