use alloc::vec;
use alloc::vec::Vec;
use ondisk_btree::ToBytes;
use vapfs::{BlockDevice, Index, Timestamp};
use crate::btree::{Directory, EntryType};
use crate::{bitmap, find_first_unallocated_inode, flush_count_entries, get_superblock, JournaledWriteResult, read_datablock, read_inode, schedule_count_assertion, schedule_data_allocation_set, schedule_multi_block_write, schedule_single_block_write, structs, why_cant_make_journal_entry, write_datablock, write_inode, write_journal_entry, write_superblock};
use crate::listblock::ListblockIter;
use crate::structs::{Filetype, Inode, JBRTargetType, JournalBlockWrite, JournalEntry, JournalEntryContents, ListBlock, Superblock, UNIXMode};

pub enum FilesystemError {
    /// Returned when an error relating to a lack of storage space occurs
    NotEnoughStorageSpace,
    /// Returned during filesystem initialisation if the block size is too small
    BlockSizeTooSmall,
    /// Returned when an error occurs that was not expected
    UnexpectedError,
    /// Returned when a provided inode is invalid
    InvalidInode,
    /// Returned when the block device is acting in an unexpected way
    UnexpectedBlockDeviceError,
    /// Returned when the filesystem may be corrupted
    PotentialFilesystemCorruption,
    /// Returned when a file is not found
    FileNotFound,
    /// Returned when a file's type is not correct for the operation
    InvalidFileType,
}

fn why(sb: &mut Superblock, bd: &mut dyn BlockDevice, res: Option<Index>) -> Result<(), FilesystemError> {
    if res.is_none() {
        let why = why_cant_make_journal_entry(sb, bd);
        match why {
            JournaledWriteResult::Success => {}
            JournaledWriteResult::OutOfDiskSpace => {
                return Err(FilesystemError::NotEnoughStorageSpace);
            }
            JournaledWriteResult::UnderlyingBlockDeviceError => {
                return Err(FilesystemError::UnexpectedBlockDeviceError);
            }
            JournaledWriteResult::PotentialFilesystemCorruption => {
                return Err(FilesystemError::PotentialFilesystemCorruption);
            }
        }
    }
    Ok(())
}

fn inode2dir(sb: &Superblock, bd: &mut dyn BlockDevice, inode_idx: Index) -> Result<Directory, FilesystemError> {
    let inode = read_inode(inode_idx, &sb, bd);
    if inode.is_none() {
        return Err(FilesystemError::InvalidInode);
    }
    let inode = inode.unwrap();
    let mut buf = vec![];
    let mut lb_iter = ListblockIter::new(&inode.listblock);
    while let Some(block) = lb_iter.next(&sb, bd) {
        let block_buf = read_datablock(block, &sb, bd);
        buf.extend_from_slice(&block_buf);
    }

    // truncate to the size of the directory
    buf.truncate(inode.size as usize);

    Ok(Directory::open(&buf))
}

/// should always return a vec aligned to the block size
fn pretend_set_allocation_inode(sb: &Superblock, bd: &mut dyn BlockDevice, index: Index, allocated: bool) -> Result<(u64, Vec<u8>), FilesystemError> {
    // todo! we should maybe optimise this to only write the byte that contains the bit, instead of the whole bitmap
    // todo! see how much time this saves?

    // inode bitmap is at 1024 + block size + datablock_bitmap_length byte offset,
    // length is inode_count / 8 rounded up
    let bitmap_offset = 1024 + sb.block_size as u64 + ((sb.data_block_count + 7) / 8);
    let bitmap_length = (sb.inode_count + 7) / 8;
    if index >= bitmap_length {
        return Err(FilesystemError::InvalidInode);
    }
    let mut bitmap_buf: Vec<u8> = Vec::new();
    bitmap_buf.resize(bitmap_length as usize, 0);
    bd.seek(bitmap_offset);
    let read_count = bd.read_blocks(&mut bitmap_buf);
    if read_count < bitmap_length as usize {
        return Err(FilesystemError::UnexpectedBlockDeviceError);
    }
    bitmap::set_bit(&mut bitmap_buf, index as usize, allocated);
    let bitmap_block_count = (bitmap_length + (sb.block_size as u64 - 1)) / sb.block_size as u64;
    let mut final_buf = vec![0; bitmap_block_count as usize * sb.block_size as usize];
    final_buf[..bitmap_buf.len()].copy_from_slice(&bitmap_buf);
    let bitmap_block_offset = bitmap_offset / sb.block_size as u64;
    Ok((bitmap_block_offset, final_buf))
}

/// initialises a filesystem with the given details
/// block_size is in bytes, and must be at least 2048 bytes
/// bd_size is in blocks, and should not exceed the size of the device that `bd` represents
/// current_time is the current time, in seconds since the unix epoch
/// inode_count is the number of inodes to create, or None to use the default (1/8 of the blocks available)
/// journal_count is the number of journal blocks to create, or None to use the default (128 blocks)
pub fn init_filesystem(bd: &mut dyn BlockDevice, block_size: u32, bd_size: u64, current_time: Timestamp, inode_count: Option<u64>, journal_count: Option<u64>) -> Result<(), FilesystemError> {
    if block_size < 2048 {
        return Err(FilesystemError::BlockSizeTooSmall);
    }
    let mut blocks_available = bd_size as i64;
    // if we have less than 5 blocks (probably around the minimum size), return not enough storage space
    if blocks_available < 5 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // minus one for the superblock
    blocks_available -= 1;
    // subtract journal count from blocks available
    blocks_available -= journal_count.unwrap_or(128) as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // inode count is the number of blocks available divided by 8
    let inode_count = inode_count.unwrap_or(blocks_available as u64 / 8);
    // each inode takes up one block, subtract that from the blocks available
    blocks_available -= inode_count as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // divide inode count by 8 (rounded up) to get the size of the inode bitmap
    let inode_bitmap_size = (inode_count + 7) / 8;
    // divide the bitmap size by the block size to get the number of blocks needed for the inode bitmap
    let inode_bitmap_blocks = (inode_bitmap_size + (block_size as u64 - 1)) / block_size as u64;
    // subtract the number of blocks needed for the inode bitmap from the blocks available
    blocks_available -= inode_bitmap_blocks as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // divide the number of blocks available by 8 to get size of the data bitmap
    let data_bitmap_blocks = (blocks_available as u64 + 7) / 8;
    // divide the data bitmap size by the block size to get the number of blocks needed for the data bitmap
    let data_bitmap_blocks = (data_bitmap_blocks + (block_size as u64 - 1)) / block_size as u64;
    // subtract the number of blocks needed for the data bitmap from the blocks available
    blocks_available -= data_bitmap_blocks as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }

    // layout of disk is roughly
    // superblock
    // data block bitmap
    // inode block bitmap
    // inode blocks
    // journal blocks
    // data blocks
    let sb_block = 1;
    let data_bitmap_block = sb_block + 1;
    let inode_bitmap_block = data_bitmap_block + data_bitmap_blocks;
    let inode_blocks = inode_bitmap_block + inode_bitmap_blocks;
    let journal_blocks = inode_blocks + inode_count;
    let data_blocks = journal_blocks + journal_count.unwrap_or(128);

    // create the superblock
    let mut sb = Superblock {
        magic: structs::MAGIC,
        block_size,
        first_data_block: data_blocks,
        first_inode_block: inode_blocks,
        first_journal_block: journal_blocks,
        inode_count,
        data_block_count: blocks_available as u64,
        journal_block_count: journal_count.unwrap_or(128),
        creation_time: current_time,
        last_modification_time: current_time,
        checksum: 0,
        journal_position: 0,
        reserved: [0; 7],
    };

    // calculate the checksum
    sb.convert_native_to_big_endian();
    sb.recalculate_checksum();
    sb.convert_big_endian_to_native();

    // create the first inode, which is the root directory
    let root_perms = UNIXMode::ReadOwner as u16 | UNIXMode::WriteOwner as u16 |
        UNIXMode::ExecuteOwner as u16 | UNIXMode::ReadGroup as u16 |
        UNIXMode::ExecuteGroup as u16 | UNIXMode::Read as u16 | UNIXMode::Execute as u16 |
        Filetype::Directory as u16;
    let mut root_inode = Inode {
        mode: root_perms,
        link_count: 0,
        uid: 0,
        gid: 0,
        size: 0,
        block_count: 0,
        creation_time: current_time,
        last_access_time: current_time,
        last_modification_time: current_time,
        last_inode_modification_time: current_time,
        deletion_time: 0,
        flags: 0,
        listblock: ListBlock {
            count: 0,
            direct_block_addresses: [0; 32],
            single_indirect_block_address: [0; 32],
            double_indirect_block_address: [0; 32],
            triple_indirect_block_address: [0; 32],
            quadruple_indirect_block_address: [0; 32],
            quintuple_indirect_block_address: [0; 32],
            sextuple_indirect_block_address: [0; 32],
        },
        checksum: 0,
    };

    let root_data = Directory::new();
    let root_bytes = root_data.to_bytes();
    // how many blocks are needed to store the root directory
    let root_block_count = (root_bytes.len() + (block_size as usize - 1)) / block_size as usize;
    // if its over 32 blocks, return unexpected error
    // (unless my b-tree implementation is awful, an empty directory should not take up 64KB)
    if root_block_count > 32 {
        return Err(FilesystemError::UnexpectedError);
    }
    let mut left = root_block_count;
    let mut buffer = vec![0; block_size as usize];
    let mut blocki = 0;
    let mut bufi = 0;
    let mut datai = 0;
    while left > 0 {
        buffer.clear();
        while bufi < block_size && datai < root_bytes.len() {
            buffer.push(root_bytes[datai]);
            bufi += 1;
            datai += 1;
        }
        unsafe { write_datablock(blocki, &sb, bd, &buffer) };
        blocki += 1;
        bufi = 0;
        left -= 1;
    }
    root_inode.size = root_bytes.len() as u64;
    root_inode.block_count = root_block_count as u64;
    for i in 0..root_block_count {
        root_inode.listblock.direct_block_addresses[i] = i as u64;
    }
    root_inode.listblock.count = root_block_count as u64;

    // recalculate the inode checksum
    root_inode.recalculate_checksum();

    // write the root inode to the first inode block
    unsafe {
        write_inode(0, &sb, bd, root_inode);
    }

    // create the inode bitmap
    let mut inode_bitmap = vec![0u8; inode_bitmap_size as usize];
    // set the first bit to 1, as the root inode is allocated
    inode_bitmap[0] |= 1;
    // write the inode bitmap to the inode bitmap block
    bd.seek(inode_bitmap_block * block_size as u64);
    bd.write_blocks(&inode_bitmap);

    // create the data bitmap
    let mut data_bitmap = vec![0u8; data_bitmap_blocks as usize];
    for i in 0..root_block_count {
        bitmap::set_bit(&mut data_bitmap, i, true);
    }
    // write the data bitmap to the data bitmap block
    bd.seek(data_bitmap_block * block_size as u64);
    bd.write_blocks(&data_bitmap);

    // zero out the journal blocks
    for i in 0..journal_blocks {
        unsafe {
            write_journal_entry(i, &sb, bd, JournalEntry {
                operation: 0,
                zeroed_content_crc32: 0,
                content: JournalEntryContents {
                    block_write: JournalBlockWrite {
                        flags: 0,
                        target_type: 0,
                        target_inode: 0,
                        target_block: 0,
                        real_target_block: 0,
                        source_block: 0,
                        source_block_crc32: 0,
                    },
                },
            });
        }
    }

    // write the superblock to the superblock block
    unsafe {
        write_superblock(sb, bd);
    }

    Ok(())
}

/// creates a new file under the given directory,
/// if initial_data is Some, the file will be filled with the given data
/// returns the inode index on success
pub fn new_file(bd: &mut dyn BlockDevice, parent_inode_idx: Index, name: &str, base_inode: Inode, initial_data: Option<&[u8]>, flush_immediately: bool) -> Result<Index, FilesystemError> {
    // we need to do the following:
    // 1. flush all journal entries to get a readable state
    // 2. create an inode (journaled block write on the inode bitmap)
    // 3. update the parent directory (journaled multiblock write on the parent directory)
    // steps 2-3 should be protected by a count assertion

    // step 1
    // flush all journal entries
    let sb = get_superblock(bd);
    if sb.is_none() {
        return Err(FilesystemError::UnexpectedError);
    }
    let mut sb = sb.unwrap();
    flush_count_entries(&sb, bd, sb.journal_position as u64, false);

    // x journal entries to allocate the inode, 1 journal entry to fill the inode, 1 journal entry to update the parent directory
    // 2 + x journal entries total

    // read the parent directory
    let mut parent_dir = inode2dir(&sb, bd, parent_inode_idx)?;

    // find a free inode
    let inode = find_first_unallocated_inode(&sb, bd);
    if inode.is_none() {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    let inode_index = inode.unwrap();
    let (bitmap_block_offset, bitmap_buf) = pretend_set_allocation_inode(&sb, bd, inode_index, true)?;

    // make a count assertion entry
    let res = schedule_count_assertion(&sb, bd, (2 + (bitmap_buf.len() / sb.block_size as usize) + if initial_data.is_some() { 1 } else { 0 }) as u32);
    why(&mut sb, bd, res)?;

    // step 2
    // journal the bitmap
    for i in 0..bitmap_buf.len() as u64 / sb.block_size as u64 {
        let res = schedule_single_block_write(
            &sb, bd,
            0,
            JBRTargetType::Disk, Some(bitmap_block_offset + i),
            &bitmap_buf[(i * sb.block_size as u64) as usize..((i + 1) * sb.block_size as u64) as usize]);
        why(&mut sb, bd, res)?;
    }

    // create the inode
    let mut inode = base_inode;
    inode.recalculate_checksum();
    inode.convert_native_to_big_endian();
    let mut buf = vec![0; sb.block_size as usize];
    unsafe { core::ptr::copy(&inode as *const Inode as *const u8, buf.as_mut_ptr(), sb.block_size as usize) };

    let res = schedule_single_block_write(
        &sb, bd,
        inode_index,
        JBRTargetType::Inode, None,
        &buf);
    why(&mut sb, bd, res)?;

    // update the parent directory
    parent_dir.new_entry(name, EntryType::Inode(inode_index));
    let mut buf = parent_dir.to_bytes();

    let res = schedule_multi_block_write(
        &sb, bd,
        parent_inode_idx,
        0, (buf.len() as Index + sb.block_size as Index - 1) / sb.block_size as Index,
        &mut buf);
    why(&mut sb, bd, res)?;
    let res = res.unwrap();

    // if flush_immediately is true, flush the journal entries
    #[allow(clippy::collapsible_if)]
    if flush_immediately {
        if !flush_count_entries(&sb, bd, res, true) {
            return Err(FilesystemError::UnexpectedError);
        }
    }

    Ok(inode_index)
}

/// removes a file from the given directory,
/// if shred is true, the file will be overwritten with zeros before being removed.
/// will only remove the file if it is a regular file or a hard link
/// returns () on success, or an error if the file could not be removed
pub fn remove_file(bd: &mut dyn BlockDevice, parent_node_idx: Index, name: &str, shred: bool, flush_immediately: bool) -> Result<(), FilesystemError> {
    // we need to do the following:
    // 1. flush all journal entries to get a readable state
    // 2. update the parent directory (journaled multiblock write on the parent directory)
    // 3. free the inode (journaled block write on the inode bitmap)
    // 4. free the data blocks (journaled data allocation set)
    // (if shred) 5. unjournaled write of zeros to the data blocks
    // steps 2-4 should be protected by a count assertion
    // 5 isn't because it's not journaled

    // flush all journal entries
    let sb = get_superblock(bd);
    if sb.is_none() {
        return Err(FilesystemError::UnexpectedError);
    }
    let mut sb = sb.unwrap();
    flush_count_entries(&sb, bd, sb.journal_position as u64, false);

    // read the parent directory
    let mut parent_dir = inode2dir(&sb, bd, parent_node_idx)?;

    // prepare to unallocate the inode
    let inode_index = parent_dir.find(name);
    let inode_index = match inode_index {
        None => {
            return Err(FilesystemError::FileNotFound);
        }
        Some(entry_type) => match entry_type {
            EntryType::Inode(inode) => inode,
            EntryType::HardLink(inode) => inode,
            EntryType::SoftLink(_) => {
                return Err(FilesystemError::InvalidFileType);
            }
            EntryType::Corrupt => {
                return Err(FilesystemError::InvalidFileType);
            }
        }
    };
    let inode = read_inode(inode_index, &sb, bd).ok_or(FilesystemError::InvalidInode)?;

    // make a count assertion entry
    // free inode = bitmap_buf.len() / sb.block_size
    // free data blocks = (inode.block_count / 32)
    // update parent directory = 1

    let res = schedule_count_assertion(&sb, bd, (2 + ((inode.block_count as usize + 31) / 32)) as u32);
    why(&mut sb, bd, res)?;

    // update the parent directory
    parent_dir.remove_entry(name);
    let mut buf = parent_dir.to_bytes();

    let res = schedule_multi_block_write(
        &sb, bd,
        parent_node_idx,
        0, (buf.len() as Index + sb.block_size as Index - 1) / sb.block_size as Index,
        &mut buf);
    why(&mut sb, bd, res)?;

    // if shred is true, overwrite the data blocks with zeros
    if shred {
        let mut lb_iter = ListblockIter::new(&inode.listblock);
        while let Some(block) = lb_iter.next(&sb, bd) {
            let buf = vec![0; sb.block_size as usize];
            unsafe {
                write_datablock(block, &sb, bd, &buf);
            }
        }
    }

    // free the inode
    let res = schedule_data_allocation_set(&sb, bd, vec![inode_index], true, false);
    why(&mut sb, bd, res)?;

    // free the data blocks
    let mut lb_iter = ListblockIter::new(&inode.listblock);
    let mut blocks = vec![];
    let mut blocks32 = vec![];
    while let Some(block) = lb_iter.next(&sb, bd) {
        blocks32.push(block);
        if blocks32.len() == 32 {
            blocks.push(blocks32);
            blocks32 = vec![];
        }
    }
    let mut last_index = 0;
    for blocks32 in blocks {
        let res = schedule_data_allocation_set(&sb, bd, blocks32, false, false);
        why(&mut sb, bd, res)?;
        last_index = res.unwrap();
    }

    // if flush_immediately is true, flush the journal entries
    #[allow(clippy::collapsible_if)]
    if flush_immediately {
        if !flush_count_entries(&sb, bd, last_index, true) {
            return Err(FilesystemError::UnexpectedError);
        }
    }

    Ok(())
}

/// writes data to a file, starting from the given byte offset
/// returns the number of bytes written on success, or an error if the file could not be written to
pub fn write_bytes_to_file(bd: &mut dyn BlockDevice, inode: Index, offset: Index, data: &[u8]) -> Result<usize, FilesystemError> {
    // we need to do the following:
    // 1. flush all journal entries to get a readable state
    // 2. calculate number of new blocks needed to store the data
    // 3. if new blocks are needed, allocate them (journaled data allocation set) and update inode with new size and block count
    // 4. write the data (un-count-asserted single block journal writes) todo: reimplement multiblock journal write so that we can use it instead


    // flush all journal entries
    let sb = get_superblock(bd);
    if sb.is_none() {
        return Err(FilesystemError::UnexpectedError);
    }
    let mut sb = sb.unwrap();
    flush_count_entries(&sb, bd, sb.journal_position as u64, false);
    
    // read the inode
    let mut inode = read_inode(inode, &sb, bd).ok_or(FilesystemError::InvalidInode)?;
    
    // calculate the number of new blocks needed to store the data
    let current_block_count = inode.block_count;
    let blocks_to_write = (data.len() as Index+ (sb.block_size as Index - 1)) / sb.block_size as Index;
    let offset_blocks = (offset + (sb.block_size as Index - 1)) / sb.block_size as Index;
    let new_block_count = if offset_blocks + blocks_to_write > current_block_count {
        offset_blocks + blocks_to_write
    } else {
        current_block_count
    };
    
    if new_block_count > current_block_count {
        // allocate new blocks and place them at the end of the listblock
    }
    todo!()
}